// torch.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <glut++.h>
#include <GL/glew.h>
#include <cmath>
#include <cstring>
#include <cstdlib>

#define MATHEPSILON 1e-10

struct _Matrix4x4{
	float mat[16];
};
typedef struct _Matrix4x4 Matrix4x4;

float determinantf (Matrix4x4 * thismatrix)
{
	return (thismatrix->mat[0] * ((thismatrix->mat[5]*thismatrix->mat[10]) - (thismatrix->mat[6] * thismatrix->mat[9]))) -
		(thismatrix->mat[1] * ((thismatrix->mat[4] * thismatrix->mat[10]) - (thismatrix->mat[6] * thismatrix->mat[8]))) +
		(thismatrix->mat[2] * ((thismatrix->mat[4] * thismatrix->mat[9]) -  (thismatrix->mat[5] * thismatrix->mat[8])));
}

void invertMatrixf (Matrix4x4 * thismatrix, Matrix4x4 * result)
{
	float fDetInverse = 1.0f / determinantf (thismatrix);
	result->mat[0]  = ( fDetInverse * ((thismatrix->mat[5] * thismatrix->mat[10]) - (thismatrix->mat[6]* thismatrix->mat[9])));
	result->mat[1]  = (-fDetInverse * ((thismatrix->mat[1] * thismatrix->mat[10]) - (thismatrix->mat[2] * thismatrix->mat[9])));
	result->mat[2]  = ( fDetInverse * ((thismatrix->mat[1] * thismatrix->mat[6])  - (thismatrix->mat[2] * thismatrix->mat[5])));
	result->mat[3]  = 0;

	result->mat[4]  = (-fDetInverse* ((thismatrix->mat[4]* thismatrix->mat[10]) - (thismatrix->mat[6]* thismatrix->mat[8])));
	result->mat[5]  = ( fDetInverse* ((thismatrix->mat[0]* thismatrix->mat[10]) - (thismatrix->mat[2]* thismatrix->mat[8])));
	result->mat[6]  = (-fDetInverse* ((thismatrix->mat[0]* thismatrix->mat[6])  - (thismatrix->mat[2]* thismatrix->mat[4])));
	result->mat[7]  = 0;

	result->mat[8]  = ( fDetInverse* ((thismatrix->mat[4]* thismatrix->mat[9]) - (thismatrix->mat[5]* thismatrix->mat[8])));
	result->mat[9]  = (-fDetInverse* ((thismatrix->mat[0]* thismatrix->mat[9]) - (thismatrix->mat[1]* thismatrix->mat[8])));
	result->mat[10] = ( fDetInverse* ((thismatrix->mat[0]* thismatrix->mat[5]) - (thismatrix->mat[1]* thismatrix->mat[4])));
	result->mat[11] = 0;

	result->mat[12] = -((thismatrix->mat[12]* result->mat[0]) + (thismatrix->mat[13]* result->mat[4]) + (thismatrix->mat[14]* result->mat[8]));
	result->mat[13] = -((thismatrix->mat[12]* result->mat[1]) + (thismatrix->mat[13]* result->mat[5]) + (thismatrix->mat[14]* result->mat[9]));
	result->mat[14] = -((thismatrix->mat[12]* result->mat[2]) + (thismatrix->mat[13]* result->mat[6]) + (thismatrix->mat[14]* result->mat[10]));
	result->mat[15] = 1.0f;
}

void normalizeVectorf (float * v)
{
	float r;
	r = (float)sqrt (v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	if (r < MATHEPSILON) return;
	r = 1.0f/r;
	v[0] = v[0]*r;
	v[1] = v[1]*r;
	v[2] = v[2]*r;
}

void crossProductf (float v0[3], float v1[3], float r[3])
{
	r[0] = v0[1] * v1[2] - v0[2] * v1[1];
	r[1] = v0[2] * v1[0] - v0[0] * v1[2];
	r[2] = v0[0] * v1[1] - v0[1] * v1[0];
}

void multMatrixVectorf (float * M,float *v)
{
	float res[3];

	res[0] = M[0]*v[0] + M[4]*v[1] + M[8]*v[2];
	res[1] = M[1]*v[0] + M[5]*v[1] + M[9]*v[2];
	res[2] = M[2]*v[0] + M[6]*v[1] + M[10]*v[2];
	
	v[0] = (float)(res[0]) + M[12];
	v[1] = (float)(res[1]) + M[13];
	v[2] = (float)(res[2]) + M[14];
}

void multMatrix33Vectorf (float *M, float *v)
{
	float res[3];

	res[0] = M[0]*v[0] + M[4]*v[1] + M[8]*v[2];
	res[1] = M[1]*v[0] + M[5]*v[1] + M[9]*v[2];
	res[2] = M[2]*v[0] + M[6]*v[1] + M[10]*v[2];
	
	v[0] = (float)(res[0]);
	v[1] = (float)(res[1]);
	v[2] = (float)(res[2]);
}

class MyApp: public GLUT::Application {
public:
	MyApp(int *ac, char **av): Application(ac, av)
	{}
};

template<typename T, unsigned int size = 3>
struct Vec {
	T coord[size];
	Vec() {
		memset(coord, 0, sizeof*coord*size);
	}
	void operator *=(float t) {
		for (int i=0; i<size; i++)
			coord[i] *= t;
	}
	void operator +=(const Vec &v) {
		for (int i=0; i<size; i++)
			coord[i] += v.coord[i];
	}
	void operator -=(const Vec &v) {
		for (int i=0; i<size; i++)
			coord[i] -= v.coord[i];
	}
};

template<typename T, unsigned int size = 4>
struct Mat {
	T elem[size*size];
	Mat() {
		memset(elem, 0, sizeof*elem*size*size);
	}
};

struct vec3f: public Vec<float> {
	vec3f(): Vec() {}
	vec3f(float x, float y, float z) {
		coord[0] = x;
		coord[1] = y;
		coord[2] = z;
	}
};
typedef Vec<float, 4> vec4f;
typedef Mat<float> mat4f;

struct Billboard {
	vec3f wcenter;
	float wwidth, wheight;
	Billboard(const vec3f &center, float width, float height):
		wcenter(center), wwidth(width), wheight(height)
	{}
	void draw() {
		calcPos();
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex3fv(coord);
		glTexCoord2f(1, 0);
		glVertex3fv(coord+3);
		glTexCoord2f(1, 1);
		glVertex3fv(coord+6);
		glTexCoord2f(0, 1);
		glVertex3fv(coord+9);
		glEnd();
	}
protected:
	float coord[12];
	void calcPos() {
		vec3f u, v, n;
		Matrix4x4 MV, invMV;
		glGetFloatv(GL_MODELVIEW_MATRIX, MV.mat);
		invertMatrixf(&MV, &invMV);

		n = wcenter;
		multMatrixVectorf(MV.mat, n.coord);
		n *= -1;

		v.coord[0] = -n.coord[2];
		v.coord[1] = 0.;
		v.coord[2] = n.coord[0];

		crossProductf(v.coord, n.coord, u.coord);

		multMatrix33Vectorf(invMV.mat, u.coord);
		multMatrix33Vectorf(invMV.mat, v.coord);
		multMatrix33Vectorf(invMV.mat, n.coord);

		normalizeVectorf(u.coord);
		normalizeVectorf(v.coord);
		normalizeVectorf(n.coord);
		u *= wheight/2.f;
		v *= wwidth/2.f;

		vec3f t = wcenter;
		t += u;
		t += v;
		memcpy(&coord[0], t.coord, sizeof*coord*3);
		t = wcenter;
		t += u;
		t -= v;
		memcpy(&coord[3], t.coord, sizeof*coord*3);
		t = wcenter;
		t -= u;
		t -= v;
		memcpy(&coord[6], t.coord, sizeof*coord*3);
		t = wcenter;
		t -= u;
		t += v;
		memcpy(&coord[9], t.coord, sizeof*coord*3);
	}
};

class MyWindow: public GLUT::Window {
public:
	void initShader() {
		const char *vertShaderSrc = "\
			void main() {\n\
			gl_Position = ftransform();\n\
				gl_TexCoord[0] = gl_MultiTexCoord0;\n\
			}\n\
		";
		const char *fragShaderSrc = "\
			uniform vec4 torchColor[4];\n\
			uniform float p[3];\n\
			varying vec2 texc;\n\
			vec4 fire(in float x0, in float y, in float pert) {\n\
			float x = x0+y*sin(sqrt(y*pert))/20.;\n\
				float t = sqrt(2.0*y)-1.;\n\
				float f = exp(-(x*x))*sqrt(y)*exp(-t*t/3.);\n\
				float g = 1.-f;\n\
				return f*f*f*torchColor[0]+ 3.*f*f*g*torchColor[1]+ 3.*f*g*g*torchColor[2]+ g*g*g*torchColor[3];\n\
			}\n\
			void main() {\n\
				float x = 12.0*(gl_TexCoord[0].s-0.5);\n\
				float y = (1.-gl_TexCoord[0].t)*12.;\n\
				gl_FragColor = (fire(x,y,p[0])+fire(x,y,p[1])+fire(x,y,p[2]))/3.;\n\
			}\n\
		";

		GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(vertShader, 1, &vertShaderSrc, NULL);
		glCompileShader(vertShader);
		GLint ret;
		glGetShaderiv(vertShader, GL_COMPILE_STATUS, &ret);
		if (!ret) {
			GLint len;
			glGetShaderiv(vertShader, GL_INFO_LOG_LENGTH, &len);
			char *buf = new char[len];
			glGetShaderInfoLog(vertShader, len, NULL, buf);
			fprintf(stderr, "Vertex Shader Compile error: %s\n", buf);
			delete [] buf;
		}
		torchProg = glCreateProgram();
		glAttachShader(torchProg, vertShader);

		GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(fragShader, 1, &fragShaderSrc, NULL);
		glCompileShader(fragShader);
		glGetShaderiv(fragShader, GL_COMPILE_STATUS, &ret);
		if (!ret) {
			GLint len;
			glGetShaderiv(fragShader, GL_INFO_LOG_LENGTH, &len);
			char *buf = new char[len];
			glGetShaderInfoLog(fragShader, len, NULL, buf);
			fprintf(stderr, "Fragment Shader Compile error: %s\n", buf);
			delete [] buf;
		}
		glAttachShader(torchProg, fragShader);

		glLinkProgram(torchProg);
		glGetProgramiv(torchProg, GL_LINK_STATUS, &ret);
		glUseProgram(torchProg);

		GLint loc = glGetUniformLocation(torchProg, "torchColor");
		float color[] = {
			1., 1., 0.5, 1.0,
			1.0, .19f, 0.0f, 1.0,
			1.0f, 1.0f, 0.1f, 1.0,
			1.0, 1.0, 1.0, 0.0
		};
		glUniform4fv(loc, 4, color);
	}
	MyWindow(int w, int h):
		Window(w, h),
		bb(vec3f(0.f, 0.f, 0.f), 0.1f, 0.1f)
	{
		glClearColor(1.f, 1.f, 1.f, 0.f);
		glClearDepth(1.f);
//		glEnable(GL_LIGHTING);
//		glEnable(GL_LIGHT0);
//		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_TEXTURE_2D);
		{GLenum err = glGetError(); if (err){printf(" t2d %x\n", err);getchar();}}
		glBindTexture(GL_TEXTURE_2D, 1);
		{GLenum err = glGetError(); if (err){printf(" bt %x\n", err);getchar();}}
		char m[4];
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		{GLenum err = glGetError(); if (err){printf(" tp %x\n", err);getchar();}}
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 1, 1, 0, GL_RGB, GL_BYTE, m);
		{GLenum err = glGetError(); if (err){printf(" ti %x\n", err);getchar();}}
//		glEnable(GL_CULL_FACE);
//		glEnable(GL_TEXTURE0);
		for (int i=0; i<sizeof p/sizeof*p; i++)
			p[i] = float(rand())*20.f/RAND_MAX;
	}
	void reshape(int w, int h) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45., double(w)/h, 0.01, 10.);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		gluLookAt(0., 0., 0.35, 0., 0., 0., 0., 1., 0.);
		glViewport(0, 0, w, h);
		redisplay();
	}
	void display() {
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glTranslated(0.05, 0.075, 0);
		for (int i=0; i<sizeof p/sizeof*p; i++)
			p[i] += (float(rand())*10.f/RAND_MAX-5.f);
		GLint loc = glGetUniformLocation(torchProg, "p");
		glUniform1fv(loc, sizeof p/sizeof*p, p);
		bb.draw();
		glPopMatrix();
		//glColor3d(0.4, 0.34, 0.22);
		//GLUquadric *q = gluNewQuadric();
		//gluQuadricDrawStyle(q, GLU_SMOOTH);
		//glPushMatrix();
		//glTranslated(0, -0.075, 0);
		//glRotated(-90, 1, 0, 0);
		//gluCylinder(q, 0.05, 0, 0.1, 16, 16);
		//glPopMatrix();
		//gluDeleteQuadric(q);
	}
	void timer(int id) {
		setTimer(1, 100);
		redisplay();
	}
	void keyboard(unsigned char key, int x, int y) {
		switch (key) {
			case 'a':
			case 'j':
				bb.wcenter.coord[0] -= 0.01f;
			break;
			case 'd':
			case 'l':
				bb.wcenter.coord[0] += 0.01f;
			break;
			case 'w':
				bb.wcenter.coord[2] -= 0.01f;
			break;
			case 's':
				bb.wcenter.coord[2] += 0.01f;
			break;
			case 'i':
				bb.wcenter.coord[1] += 0.01f;
			break;
			case 'k':
				bb.wcenter.coord[1] -= 0.01f;
			break;
			default:
				GLUT::Window::keyboard(key, x, y);
		}
		redisplay();
	}
protected:
	Billboard bb;
	float p[3];
	GLuint torchProg;
};

int main(int argc, char* argv[])
{
	MyApp app(&argc, argv);
	MyWindow win(800, 600);
	win.setTitle("Torch demo");
	glewInit();
	win.initShader();
	win.setTimer(1, 50);
	app.start();
	return 0;
}
